\spiro.inc	6-Oct-2015
\Graphic and mouse routines for spiro.xpl

string	0;			\null-terminated strings are the standard here

def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,  \attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \EGA palette
def	Pi = 3.14159265358979323846;



proc	Exit(Str);		\Display error message and exit program
char	Str;
[SetVid($03);			\make sure screen is in normal text mode
Text(0, Str);  CrLf(0);
exit;
];	\Exit



func	IntLen(N);		\Return number of digits in N, a decimal integer
int	N;			\Adds 1 for a minus sign if N is negative
int	I;
for I:= if N>=0 then 1 else 2 to 20 do
	[N:= N/10;  if N=0 then return I];



proc	JustOut(Dev, N, Places, Fill); \Output a right-justified integer
\Always outputs correct value of N regardless of Places
int	Dev,	\output device number
	N,	\integer to output
	Places,	\size of field in characters (right-justifies)
	Fill;	\character to fill background of field. Usually space (Sp) or ^0
int	I;	\ Negative numbers should be filled with spaces.
[for I:= 1 to Places-IntLen(N) do ChOut(Dev, Fill);
IntOut(Dev, N);
];	\JustOut



int	PenX, PenY,		\pen pixel coordinates for DrawLine & DrawChar
	LineWidth,		\DrawLine thickness: 0 or 1 for one-pixel width
	DotDash,		\byte pattern for DrawLine: 0 = solid line
	CharDir;		\character direction: 0=hoiz; not 0 = vertical



proc	InitLib;
[PenX:= 0;  PenY:= 0;
LineWidth:= 0;
DotDash:= 0;
CharDir:= 0;
];	\InitLib



proc	MovePen(X, Y);		\Position drawing pen
int	X, Y;
[PenX:= X;  PenY:= Y];



proc	DrawChar(Font, Ch, Color);	\Display a character at PenX,PenY
int	Font, Ch, Color;		\background is transparent
int	I, J, Byte;			\chars can be rotated by 90 degrees
int	FontHi;				\char height in pixels = number of bytes
char	FontOff;			\address of table (offset)
[case Font of
  0:	[FontOff:= GetFont(0);  FontHi:= 16];
  1:	[FontOff:= GetFont(1);  FontHi:= 8];
  2:	[FontOff:= GetFont(2);  FontHi:= 14]
other	[];

for J:= 0 to FontHi-1 do
	[Byte:= FontOff(Ch*FontHi + J);
	for I:= 0 to 7 do		\for each bit in the Byte...
		[if Byte & $80 then
			if CharDir then Point(PenX+J, PenY-I, Color)
			       else Point(PenX+I, PenY+J, Color);
		Byte:= Byte << 1;
		];
	];
if CharDir then PenY:= PenY - 8
       else PenX:= PenX + 8;		\move to next character position
];	\DrawChar



proc	DrawText(Font, Str, Color);	\Display a text string at PenX,PenY
int	Font;
char	Str;
int	Color;
while Str(0) do [DrawChar(Font, Str(0), Color);  Str:= Str+1];	



proc	DrawDot(X0, Y0, C, S);	\Draw a dot centered at X0,Y0
int	X0, Y0, C, S;		\color, size
int	X, Y, X2, Y2, R2;
[if S <= 1 then [Point(X0, Y0, C);  return];
R2:= (S*S)>>1 - 1;
if R2 < 0 then R2:= 0;
for Y:= -S to +S do
	[Y2:= Y*Y;
	for X:= -S to +S do
		[X2:= X*X + Y2;
		if X2 <= R2 then
			Point(X+X0, Y+Y0, C);
		];
	];
];	\DrawDot



proc	DrawLine(X, Y, C);	\Draw line from PenX,PenY to X,Y in color C
int	X, Y, C;	\coordinates of end-of-line, color, width
int	DX, DY,		\delta X, delta Y
	IX, IY,		\plotting increments along X and Y axes
	SY,		\flag: step along Y direction
	D,		\decision variable: move straight or diagonally
	E, NE,		\East, Northeast (straight or diagonal)
	T, I;		\scratch
[DX:= X - PenX;				\delta X
if DX >= 0 then				\if positive then increment
	IX:= 1
else	[DX:= -DX;  IX:= -1];		\else absolute value and decrement

DY:= Y - PenY;				\same for Y direction
if DY >= 0 then
	IY:= 1
else	[DY:= -DY;  IY:= -1];

if DX >= DY then			\step along X direction
	SY:= false
else	[SY:= true;			\step along Y direction
	T:= DX;  DX:= DY;  DY:= T;	\swap DX and DY
	];

D:= 2*DY - DX;				\set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

DotDash:= DotDash<<1;			\rotate byte left
if DotDash & $100 then DotDash:= DotDash+1
else	DrawDot(PenX, PenY, C, LineWidth);
for I:= 1 to DX do
	[if D < 0 then
		[D:= D + E;
		if SY then PenY:= PenY + IY
		      else PenX:= PenX + IX;
		]
	else	[D:= D + NE;		\move diagonally
		PenX:= PenX + IX;
		PenY:= PenY + IY;
		];
	DotDash:= DotDash<<1;			\rotate byte left
	if DotDash & $100 then DotDash:= DotDash+1
	else	DrawDot(PenX, PenY, C, LineWidth);
	];
];	\DrawLine



proc	DrawRectangle(X0, Y0, W, H, Color, Fill); \Display a rectangle
int	X0, Y0,		\coordinates of upper-left corner
	W, H,		\width and height (including outlines)
	Color, Fill;	\flag: solid fill vs. outlined
int	I, J;
[W:= W-1;  H:= H-1;	\dimensions include outline borders
if not Fill then
	[MovePen(X0, Y0);
	DrawLine(X0+W, Y0, Color);
	DrawLine(X0+W, Y0+H, Color);
	DrawLine(X0, Y0+H, Color);
	DrawLine(X0, Y0, Color);
	]
else	[if W < 16 then				\if fill with tall and skinny
		for I:= X0 to X0+W do		\ then use vertical lines
			[Move(I, Y0);  Line(I, Y0+H, Color)]
	else	for J:= Y0 to Y0+H do		\ else use fast horizontal lines
			[Move(X0, J);  Line(X0+W, J, Color)];
	];
];	\DrawRectangle



proc	DrawSprite(X0, Y0, Spr, Trans);	\Display a sprite image
int	X0, Y0;	\coordinates of upper-left corner (pixels)
int	Spr;	\address of sprite data
int	Trans;	\flag: treat background (0=black) as transparent
int	X, Y, I, C, W, H;
[W:= Spr(0);		\get width and height (in pixels)
 H:= Spr(1);
I:= 2;
for Y:= Y0 to Y0+H-1 do
    for X:= X0 to X0+W-1 do
	[C:= Spr(I);		\get pixel's color
	if C ! ~Trans then	\background (0) can be transparent
		Point(X, Y, C);
	I:= I+1;
	];
];	\DrawSprite



proc	GetSprite(X0, Y0, W, H, Spr);	\Load sprite array from screen image
int	X0, Y0,	\coordinates on screen of upper-left corner of image
	W, H;	\width and height of sprite image (pixels)
int	Spr;	\address of array in which to load sprite data: Spr(2+W*H)
int	I, X, Y;
[Spr(0):= W;
 Spr(1):= H;
I:= 2;
for Y:= Y0 to Y0+H-1 do
    for X:= X0 to X0+W-1 do
	[Spr(I):= ReadPix(X, Y);
	I:= I+1;
	];
];	\GetSprite

\------------------------------- MOUSE ROUTINES --------------------------------

int	MouseX, MouseY,		\current mouse coordinates (pixels)
	MouseX0, MouseY0,	\previous mouse coordinates (for backgnd image)
	MouseB;			\mouse buttons: bit 0 set = left down, 1 = right
def	MouseW=12, MouseH=21;	\mouse pointer width and height (pixels)
int	MouseBg(2+MouseW*MouseH); \sprite: mouse pointer's background image



proc	DrawPointer(X, Y);	\Display mouse pointer at screen coordinates X,Y
int	X, Y;			\coordinates of upper-left corner of arrow
char	Ptr;
int	I, J;
[Ptr:= ["#           ",
	"##          ",
	"#.#         ",
	"#..#        ",
	"#...#       ",
	"#....#      ",
	"#.....#     ",
	"#......#    ",
	"#.......#   ",
	"#........#  ",
	"#.........# ",
	"#......#####",
	"#...#..#    ",
	"#..##..#    ",
	"#.#  #..#   ",
	"##   #..#   ",
	"#     #..#  ",
	"      #..#  ",
	"       #..# ",
	"       #..# ",
	"        ##  "];

for J:= 0 to MouseH-1 do		\for all the rows...
    for I:= 0 to MouseW-1 do		\  for all the pixels in the row...
	case Ptr(J,I) & $7F of
	  ^#:	Point(I+X, J+Y, 0);	\black for all video modes
	  ^.:	Point(I+X, J+Y, BWhite)
	other	[];
];	\DrawPointer



proc	ShowMouse(On);		\Turn mouse pointer on or off
int	On;	\flag: true = pointer on; false = pointer off
[if On then
	[MouseX0:= MouseX;  MouseY0:= MouseY;	\record background's coordinates
	GetSprite(MouseX0, MouseY0, MouseW, MouseH, MouseBg); \save background
	DrawPointer(MouseX, MouseY);		\show mouse pointer
	]
else	DrawSprite(MouseX0, MouseY0, MouseBg, false);	\restore backgnd image
];	\ShowMouse



proc	MoveMouse;		\Move mouse pointer
int	Mouse;
[Mouse:= GetMouse;
MouseX:= Mouse(0);  MouseY:= Mouse(1);  MouseB:= Mouse(2);
if MouseX#MouseX0 ! MouseY#MouseY0 then			\mouse moved
	[DrawSprite(MouseX0, MouseY0, MouseBg, false);	\restore backgnd image
	MouseX0:= MouseX;  MouseY0:= MouseY;		\new backgnd location
	GetSprite(MouseX0, MouseY0, MouseW, MouseH, MouseBg); \save new backgnd
	DrawPointer(MouseX, MouseY);			\show pointer at new loc
	];
];	\MoveMouse

