\spiro.xpl	6-Oct-2015	Loren Blaney		loren.blaney@gmail.com
\Spirograph maker, translated from Larry Fish's Windows version.

include spiro.inc;		\graphic and mouse routines

\Scrollbar routines:
\
\      X,Y
\	+--------------------------------------+  +---+  -
\	|16|        |16|                    |16|  |123|  H=16
\	+--------------------------------------+  +---+  -
\	|----------------- W ------------------|16| 24|
\
\ The left and right buttons and the knob are each 16 pixels wide.
\ X,Y is the coordinate of the upper-left corner of the scrollbar.
\ W is the width of the scrollbar (in pixels), and H is the height.
\ There is a 2-pixel border that extends beyond this area.
\ The knob's travel range = W - 3*16.

\This code assumes 8 bits per pixel (because of values used for colors)
def	VMode=$105,		\graphic video mode
	ScrW=1024, ScrH=768;	\screen dimensions (pixels)
\def	VMode=$101,
\	ScrW=640, ScrH=480;

int	Scrollbars,		\number of scrollbars
\Constant arrays containing data for each scrollbar:
	BarX, BarY,		\coordinates of upper-left corner (pixels)
	BarW,			\width of scrollbar including buttons (pixels)
	BarL, BarH,		\low and high limit value
	BarV,			\value, as displayed
	BarP;			\position of knob in scrollbar
real	BarF;			\conversion factor between position and value

\ F = W / (H-L)			W = BarW - 3*16
\ P = F * (V-L)
\ V = P/F + L



proc	DrawHeader(X, Y, W, H, Title);	\Draw a blue panel (blistered outward)
int	X, Y, W, H, Title;		\upper-left corner, width and height
[DrawRectangle(X, Y, W-1, H-1, $36, true);
Move(X+1,   Y+1);    Line(X+W-3, Y+1,   $35);
Move(X+1,   Y+1);    Line(X+1,   Y+H-3, $35);
Move(X+W-2, Y+H-2);  Line(X+1,   Y+H-2, $37);
Move(X+W-2, Y+H-2);  Line(X+W-2, Y+1,   $37);
Move(X+W-1, Y+H-1);  Line(X,     Y+H-1, $37);
Move(X+W-1, Y+H-1);  Line(X+W-1, Y,     $37);
MovePen(X+16+1, Y+4+1);  DrawText(0, Title, 0);		\drop shadow
MovePen(X+16,   Y+4);    DrawText(0, Title, $0F);
];	\DrawHeader



proc	DrawPanel(X, Y, W, H);		\Draw a gray panel (blistered outward)
int	X, Y, W, H;			\upper-left corner, width and height
[DrawRectangle(X, Y, W-1, H-1, $1C, true);
Move(X+1,   Y+1);    Line(X+W-3, Y+1,   $1F);
Move(X+1,   Y+1);    Line(X+1,   Y+H-3, $1F);
Move(X+W-2, Y+H-2);  Line(X+1,   Y+H-2, $18);
Move(X+W-2, Y+H-2);  Line(X+W-2, Y+1,   $18);
Move(X+W-1, Y+H-1);  Line(X,     Y+H-1, $14);
Move(X+W-1, Y+H-1);  Line(X+W-1, Y,     $14);
];	\DrawPanel



proc	DrawFrame(X, Y, W, H, Title);	\Draw a shadowed frame
int	X, Y, W, H, Title;
[DrawRectangle(X+1, Y+1, W, H, $1F, false);
 DrawRectangle(X, Y, W, H, $18, false);
 DrawRectangle(X+14, Y, 2, 2, $1C, true);
Move(X+16, Y-5);  Attrib($1C00);  Text($206, Title);
];	\DrawFrame



proc	DrawCheckeredRectangle(X, Y, W, H);	\Draw background for a scrollbar
int	X, Y, W, H;	\upper left corner coordinates, width, height (pixels)
int	J;
[for J:= Y to Y+H-1 do
	[Move(X, J);  Line(X+W-1, J, if J&1 then $AA1F else $551F);
	 Move(X, J);  Line(X+W-1, J, if J&1 then $551C else $AA1C);
	];
];	\DrawCheckeredRectangle



proc	DrawLeftArrow(X, Y);		\Draw left arrow symbol
int	X, Y;				\coordinates of upper-left corner of
[Move(X+6, Y+7);  Line(X+6, Y+7, 0);	\ button containing the arrow
 Move(X+7, Y+6);  Line(X+7, Y+8, 0);
 Move(X+8, Y+5);  Line(X+8, Y+9, 0);
 Move(X+9, Y+4);  Line(X+9, Y+10, 0);
];	\DrawLeftArrow



proc	DrawRightArrow(X, Y);		\Draw right arrow symbol
int	X, Y;				\coordinates of upper-left corner of
[Move(X+6, Y+4);  Line(X+6, Y+10, 0);	\ button containing the arrow
 Move(X+7, Y+5);  Line(X+7, Y+9, 0);
 Move(X+8, Y+6);  Line(X+8, Y+8, 0);
 Move(X+9, Y+7);  Line(X+9, Y+7, 0);
];	\DrawRightArrow



proc	DrawUpButton(X, Y);		\Draw button in its up position
int	X, Y;				\coordinates of upper-left corner
DrawPanel(X, Y, 16, 16);



proc	DrawDownButton(X, Y);		\Draw button in its down position
int	X, Y;				\coordinates of upper-left corner
[DrawRectangle(X, Y, 15, 15, $1C, true);
Move(X+ 0, Y+ 0);  Line(X+14, Y+ 0, $18);
Move(X+ 0, Y+ 0);  Line(X+ 0, Y+14, $18);
Move(X+15, Y+15);  Line(X+ 0, Y+15, $18);
Move(X+15, Y+15);  Line(X+15, Y+ 0, $18);
];	\DrawDownButton



proc	DrawWindow(X, Y, W, H);		\Draw a recessed window
int	X, Y,	\coordinates of upper-left corner of window surrounded by frame
	W, H;	\width, height
[DrawRectangle(X, Y, W, H, $1F, true);
Move(X-1, Y-1);  Line(X+W-1, Y-1, $14);	\draw shaded frame edges
Move(X-1, Y-1);  Line(X-1, Y+H-1, $14);
Move(X-2, Y-2);  Line(X+W, Y-2, $18);
Move(X-2, Y-2);  Line(X-2, Y+H, $18);
Move(X+W, Y+H);  Line(X-1, Y+H, $1C);
Move(X+W, Y+H);  Line(X+W, Y-1, $1C);
Move(X+W+1, Y+H+1);  Line(X-1, Y+H+1, $1F);
Move(X+W+1, Y+H+1);  Line(X+W+1, Y-1, $1F);
];	\DrawWindow



proc	NumOut(X, Y, N);		\Display right-justified 3-digit number
int	X, Y, N;			\upper-left coords of window, number
[Attrib($1F00);				\display black on bright white
Move(X, Y+2);
JustOut($206, N, 3, ^ );
];	\NumOut



proc	DrawKnob(I);			\Draw knob and its position value
int	I;				\index of scrollbar to display
int	X, Y, W, P;
[X:= BarX(I);  Y:= BarY(I);  W:= BarW(I);  P:= BarP(I);
DrawCheckeredRectangle(X+16, Y, W-32, 16);	\overwrite knob at old position
DrawUpButton(X+16+P, Y);			\draw knob at current position
NumOut(X+W+16, Y, BarV(I));			\display position's value
];	\DrawKnob



proc	DrawScrollbar(I);		\Draw scrollbar, incl. knob and value
int	I;				\scrollbar number
int	X, Y, W;
[X:= BarX(I);  Y:= BarY(I);  W:= BarW(I);
DrawWindow(X, Y, W, 16);		\draw main scrollbar window
DrawUpButton(X, Y);			\draw arrow buttons
DrawLeftArrow(X, Y);
DrawUpButton(X+W-16, Y);
DrawRightArrow(X+W-16, Y);
DrawWindow(X+W+15, Y, 3*8+1, 16);	\draw window for position value
DrawKnob(I);				\draw knob and its position value
];	\DrawScrollbar

\-------------------------------------------------------------------------------

proc	DrawExitX(X, Y);		\Draw X for exit button
int	X, Y;				\coordinates of upper-left corner of
[Move(X+5, Y+4);  Line(X+11, Y+10, 0);	\ button containing the X
 Move(X+4, Y+4);  Line(X+11, Y+11, 0);
 Move(X+4, Y+5);  Line(X+10, Y+11, 0);
 Move(X+4, Y+10); Line(X+10, Y+4, 0);
 Move(X+4, Y+11); Line(X+11, Y+4, 0);
 Move(X+5, Y+11); Line(X+11, Y+5, 0);
];	\DrawExitX



int	MXButton,		\flag: mouse button just got pressed for exit
	ExitPending,		\flag: mouse button went down on exit button
	ExitDown;		\flag: exit button is shown in down position
def	ExitX=358, ExitY=4;	\position of exit button



func	ExitButton;			\Return 'true' if exit button is clicked
[if MouseB & 1 then
	[if ~MXButton then		\left mouse button just went down
		[MXButton:= true;
		if MouseX>=ExitX & MouseX<ExitX+16 &
		   MouseY>=ExitY & MouseY<ExitY+16 then \exit button is pressed
			ExitPending:= true;
		];
	\If mouse is on exit button then show exit button down else show it up
	if ExitPending then
		[if MouseX>=ExitX & MouseX<ExitX+16 &	\mouse is on exit button
		   MouseY>=ExitY & MouseY<ExitY+16 then
			[if not ExitDown then
				[ShowMouse(false);
				DrawDownButton(ExitX, ExitY);
				DrawExitX(ExitX+1, ExitY+1);
				ShowMouse(true);
				ExitDown:= true;
				];
			]
		else	[			\mouse slid off exit button
			if ExitDown then
				[ShowMouse(false);
				DrawUpButton(ExitX, ExitY);
				DrawExitX(ExitX, ExitY);
				ShowMouse(true);
				ExitDown:= false;
				];
			];
		];
	]
\If exit was pending and mouse button is released while on exit button then exit
\i.e: if mouse button is released and exit button is down then exit program
else	[if ExitDown then return true;
	MXButton:= false;
	ExitPending:= false;
	];
return false;
];	\ExitButton

\-------------------------------------------------------------------------------

int	Knob,			\index (scrollbar no.) of grabbed & dragged knob
	MButton,		\flag: mouse button just got pressed
	OffsetX;		\X offset (in pixels) of knob from mouse pointer
				\ when it was grabbed


proc	MoveKnobs;			\Drag slider knobs with mouse
int	KX, KY,		\knob coordinates (and offset)
	I, Val;		\index, BarV temporary
[MoveMouse;
if MouseB & 1 then
	[if ~MButton then			\left button just went down
		[MButton:= true;
		Knob:= -1;			\assume no knob is grabbed
		for I:= 0 to Scrollbars-1 do	\check all knobs to see if one
		    [KX:= BarX(I) + 16 + BarP(I);\ is grabbed
		     KY:= BarY(I);
		     if MouseX>=KX & MouseX<KX+16 &
		        MouseY>=KY & MouseY<KY+16 then \knob is grabbed
			 [Knob:= I;		\get index of its scrollbar
			 OffsetX:= MouseX - KX;	\get X distance from knob to
			 ];			\ mouse pointer
		    ];
		];

	\Mouse button is still held down
	if Knob >= 0 then			\a knob was grabbed, so drag it
		[KX:= MouseX - OffsetX;		\get X coord of dragged knob
		 KX:= KX - BarX(Knob) - 16;	\ relative to its slot
		Val:= fix(float(KX) / BarF(Knob)) + BarL(Knob);
		if Val < BarL(Knob) then Val:= BarL(Knob);
		if Val > BarH(Knob) then Val:= BarH(Knob);
		\(Recalculating knob position provides detent action if needed)
		KX:= fix(float(Val - BarL(Knob)) * BarF(Knob));
		BarV(Knob):= Val;
		if KX # BarP(Knob) then		\knob moved so
			[BarP(Knob):= KX;	\update its new position
			ShowMouse(false);
			DrawKnob(Knob);		\and draw it there
			ShowMouse(true);
			];
		];
	]
else	[MButton:= false;			\mouse button is released
	Knob:= -1;				\release any grabbed knob
	];
];	\MoveKnobs

\-------------------------------------------------------------------------------

int	ButX, ButY,		\coordinates of pressed arrow button
	ButLeft,		\flag: pressed button points left (else right)
	Drawn,			\Up or Down state currently shown for button
	M2Button,		\flag: mouse button just got pressed
	AutoRepeat,		\flag: auto-repeat for arrow buttons
	Timer,			\count-down timer (1/18th seconds)
	Tick, Tick0;		\tick (decrements every 1/18th second)
def \Drawn\ Up, Down;



proc	PressButtons;			\Press arrow buttons with mouse
int	I,		\index
	BX, BY;		\arrow button coordinates
[if MouseB & 1 then
	[if ~M2Button then			\mouse button just went down
		[M2Button:= true;
		for I:= 0 to  Scrollbars-1 do	      \check all arrow buttons
		    [BX:= BarX(I);  BY:= BarY(I);
		    if MouseX>=BX & MouseX<BX+16 &
		       MouseY>=BY & MouseY<BY+16 then \left arrow button pressed
			[ButX:= BX;   ButY:= BY;      \save its coordinates
			ButLeft:= true;
			if not AutoRepeat then Timer:= 9; \delay before auto-
			Drawn:= Down;			  \ repeat begins
			ShowMouse(false);	\show button in down position
			DrawDownButton(ButX, ButY);
			DrawLeftArrow(ButX+1, ButY+1);
			BarV(I):= BarV(I)-1;
			if BarV(I) < BarL(I) then BarV(I):= BarL(I);
			BarP(I):= fix(float(BarV(I) - BarL(I)) * BarF(I));
			DrawKnob(I);		\show knob in its new location
			ShowMouse(true);
			];
		    BX:= BarX(I) + BarW(I) - 16;
		    if MouseX>=BX & MouseX<BX+16 &
		       MouseY>=BY & MouseY<BY+16 then \right arrow button pres'd
			[ButX:= BX;   ButY:= BY;      \save its coordinates
			ButLeft:= false;
			if not AutoRepeat then Timer:= 9; \delay before auto-
			Drawn:= Down;			  \ repeat begins
			ShowMouse(false);	\show button in down position
			DrawDownButton(ButX, ButY);
			DrawRightArrow(ButX+1, ButY+1);
			BarV(I):= BarV(I)+1;
			if BarV(I) > BarH(I) then BarV(I):= BarH(I);
			BarP(I):= fix(float(BarV(I) - BarL(I)) * BarF(I));
			DrawKnob(I);		\show knob in its new location
			ShowMouse(true);
			];
		    ]; \for
		];

	\Mouse button is still held down
	if ButX >= 0 then			\an arrow button is pressed
		[if MouseX>=ButX & MouseX<ButX+16 &	\ & mouse is still on it
		    MouseY>=ButY & MouseY<ButY+16 then
			[if Timer = 0 then	   \& time delay is expired
				[M2Button:= false; \enable arrow button to be
				AutoRepeat:= true; \ pressed again
				Timer:= 1;	   \delay for auto-repeat rate
				];
			if Drawn # Down then	\show button in down position
				[ShowMouse(false);
				DrawDownButton(ButX, ButY);
				if ButLeft then DrawLeftArrow(ButX+1, ButY+1)
					   else DrawRightArrow(ButX+1, ButY+1);
				ShowMouse(true);
				Drawn:= Down;
				];
			]
		else	[			\mouse slid off the button
			if Drawn # Up then	\show button in its up position
				[ShowMouse(false);
				DrawUpButton(ButX, ButY);
				if ButLeft then DrawLeftArrow(ButX, ButY)
					   else DrawRightArrow(ButX, ButY);
				ShowMouse(true);
				Drawn:= Up;
				];
			];
		];
	]
else	[M2Button:= false;			\mouse button is released
	AutoRepeat:= false;
	if ButX >= 0 then			\an arrow button was pressed
		[ShowMouse(false);		\ so show it in its up position
		DrawUpButton(ButX, ButY);
		if ButLeft then DrawLeftArrow(ButX, ButY)
			   else DrawRightArrow(ButX, ButY);
		ShowMouse(true);
		ButX:= -1;
		];
	];

Tick:= GetTime/55555;				\tick the count-down timer
if Tick # Tick0 then				\ every 1/18th second
	[Tick0:= Tick;
	if Timer > 0 then Timer:= Timer-1;
	];
];	\PressButtons

\-------------------------------------------------------------------------------

proc	DrawSpiro;			\Display a spirograph
real	R0, R1, R,	\radii of stator and rotor wheels, and their sum
	R2,		\pen offset from rotor wheel
	A0,		\angular position of center of rotor (radians)
	A1,		\angle of rotation of rotor (radians)
	A1Max,		\maximum rotation of rotor (limit in radians)
	C1x, C1y,	\coordinates of center of rotor wheel
	P1x, P1y;	\pen location from rotor wheel relative to C1
int	First, X, Y;
def	C0y=(ScrW-380)/2, C0x=C0y+380;	\location of center of stator on screen
[R0:= float(BarV(0));
 R1:= float(BarV(1));
 R2:= float(BarV(2));
 R:= R0 + R1;
A1Max:= float(BarV(3)) * 2.0*Pi;
A1:= 0.0;
First:= true;
repeat	\anglular position of center of rotor due to rolling inside stator
	if R # 0.0 then A0:= A1*R1/R;
	C1x:= R*Cos(A0);	\center of rotor relative to C0
	C1y:= R*Sin(A0);
	P1x:= (R1+R2)*Cos(A1);	\pen location relative to C1
	P1y:= (R1+R2)*Sin(A1);
	X:= fix(C1x + P1x);	\add vectors
	Y:= fix(C1y + P1y);
	if First then [Move(X+C0x, Y+C0y);  First:= false]
		 else Line(X+C0x, Y+C0y, LCyan);
	A1:= A1 + 0.05;		\rotate rotor about 2 degrees, for smooth curves
until	A1 >= A1Max;
];	\DrawSpiro

\-------------------------------------------------------------------------------

proc	DoScrollbars;			\Set up scrollbars and move sliders
int	I, V, V0;
[Scrollbars:= 4;				\number of scrollbars
BarX:= [16, 16, 16, 16];			\X coordinates
BarY:= [4*16, 8*16, 12*16, 16*16];		\Y coordinates
BarW:= [256+16*3, 256+16*3, 256+16*3, 256+16*3]; \widths (pixels)
BarL:= [0,  -50,   0,   0];			\low limit values
BarH:= [100, 50, 100, 255];			\high limit values
BarV:= [80,  -3,  80,  50];			\initial values
BarP:= [0, 0, 0, 0];				\positions of knobs
BarF:= [0.0, 0.0, 0.0, 0.0];			\conversion factors
for I:= 0 to  Scrollbars-1 do
	[BarF(I):= float(BarW(I)-16*3) / (float(BarH(I)-BarL(I)));
	 BarP(I):= fix(float(BarV(I) - BarL(I)) * BarF(I));
	];
DrawPanel(0, 0, 380, ScrH);			\display scrollbars
for I:= 0 to  Scrollbars-1 do DrawScrollbar(I);
DrawRectangle(0, 0, 380-1, 24, $09, true);
DrawHeader(0, 0, 380-1, 24, "Spirograph");
DrawUpButton(ExitX, ExitY);
DrawExitX(ExitX, ExitY);

DrawFrame(4, 4*16-14,  370, 44, "Fixed circle radius");
DrawFrame(4, 8*16-14,  370, 44, "Rolling circle radius");
DrawFrame(4, 12*16-14, 370, 44, "Circle offset");
DrawFrame(4, 16*16-14, 370, 44, "Revolutions");

MButton:= false;				\assume mouse button is up
Knob:= -1;					\nothing is grabbed yet
ButX:= -1;					\no arrow buttons are pressed
AutoRepeat:= false;
MXButton:= false;
ExitPending:= false;
ExitDown:= false;

ShowMouse(true);				\turn on mouse pointer

V0:= -1;	\force initial drawing to occur
repeat	MoveKnobs;
	PressButtons;
	V:= BarV(0) + BarV(1) + BarV(2) + BarV(3);
	if V # V0 then
		[V0:= V;
		DrawRectangle(380, 0, ScrW-380, ScrH, 0, true);	\erase
		DrawSpiro;
		];
until	KeyHit or ExitButton;
OpenI(0);	\eat keystroke
];	\DoScrollbars

\===============================================================================

int	Mouse;
[InitLib;
SetVid(VMode);
if not OpenMouse then Exit("This program requires a mouse.");
Mouse:= GetMouse;
MouseX:= Mouse(0);  MouseY:= Mouse(1);  MouseB:= Mouse(2);

DoScrollbars;
SetVid(3);	\restore normal text mode
]
